@@ -1,6 +1,5 @@
 #include "qap_decode.h"
 
-#include <Rversion.h>
 #include <string.h>
 
 #define decode_to_SEXP QAP_decode 
@@ -14,6 +13,14 @@ extern cetype_t string_encoding;
 #define mkRChar(X) mkCharCE((X), string_encoding)
 #endif
 
+/* this is only used in debugging output for down-casing
+   pointers to long on Windows as intermediate step */
+#ifdef WIN64
+typedef unsigned long long temp_ptr_int_t;
+#else
+typedef unsigned long temp_ptr_int_t;
+#endif
+
 /* this is the representation of NAs in strings. We chose 0xff since that should never occur in UTF-8 strings. If 0xff occurs in the beginning of a string anyway, it will be doubled to avoid misrepresentation. */
 static const unsigned char NaStringRepresentation[2] = { 255, 0 };
 
@@ -29,7 +36,7 @@ SEXP decode_to_SEXP(unsigned int **buf)
     SEXP val = 0, vatt = 0;
     int ty = PAR_TYPE(ptoi(*b));
     rlen_t ln = PAR_LEN(ptoi(*b));
-    R_len_t i, l;
+    rlen_t i, l;
     
     if (IS_LARGE(ty)) {
 	ty ^= XT_LARGE;
@@ -119,7 +126,7 @@ SEXP decode_to_SEXP(unsigned int **buf)
 	l = ln / 16;
 	val = allocVector(CPLXSXP, l);
 #ifdef NATIVE_COPY
-	memcpy(COMPLEX(val), b, sizeof(*COMPLEX(val)) * l);
+	memcpy(COMPLEX(val), b, sizeof(Rcomplex) * l);
 	b += l * 4;
 #else
 	i = 0;
@@ -176,28 +183,31 @@ SEXP decode_to_SEXP(unsigned int **buf)
     case XT_VECTOR_EXP:
 	{
 	    unsigned char *ie = (unsigned char*) b + ln;
-	    R_len_t n = 0;
-	    SEXP lh = R_NilValue;
-	    SEXP vr = allocVector(VECSXP, 1);
+	    rlen_t n = 0;
+	    SEXP vr = PROTECT(CONS(R_NilValue, R_NilValue));
+	    SEXP tail = vr;
 	    *buf = b;
-	    PROTECT(vr);
 	    while ((unsigned char*)*buf < ie) {
 		SEXP v = decode_to_SEXP(buf);
-		lh = CONS(v, lh);
-		SET_VECTOR_ELT(vr, 0, lh); /* this is our way of staying protected .. maybe not optimal .. */
+		/* CONS() protects its arguments if GC is needed so v is safe */
+		SEXP ne = CONS(v, R_NilValue);
+		SETCDR(tail, ne);
+		tail = ne;
 		n++;
 	    }
 #ifdef RSERV_DEBUG
-	    printf(" vector (%s), %d elements\n", (ty == XT_VECTOR) ? "generic" : ((ty == XT_VECTOR_EXP) ? "expression" : "string"), n);
+	    printf(" vector (%s), %ld elements\n", (ty == XT_VECTOR) ? "generic" : ((ty == XT_VECTOR_EXP) ? "expression" : "string"), (long) n);
 #endif
 	    val = PROTECT(allocVector((ty==XT_VECTOR) ? VECSXP : ((ty == XT_VECTOR_EXP) ? EXPRSXP : STRSXP), n));
-	    while (n > 0) {
-		n--;
-		SET_VECTOR_ELT(val, n, CAR(lh));
-		lh = CDR(lh);
+	    n = 0;
+	    while (CDR(vr) != R_NilValue) {
+		vr = CDR(vr);
+		SET_VECTOR_ELT(val, n++, CAR(vr));
 	    }
 #ifdef RSERV_DEBUG
-	    printf(" end of vector %lx/%lx\n", (long) *buf, (long) ie);
+	    printf(" end of vector %lx/%lx\n",
+		   (unsigned long) ((temp_ptr_int_t) *buf),
+		   (unsigned long) ((temp_ptr_int_t) ie));
 #endif
 	    UNPROTECT(2); /* val and vr */
 	    break;
@@ -235,13 +245,17 @@ SEXP decode_to_SEXP(unsigned int **buf)
 	    *buf = b;
 	    while ((unsigned char*)*buf < ie) {
 #ifdef RSERV_DEBUG
-		printf(" el %08lx of %08lx\n", (unsigned long)*buf, (unsigned long) ie);
+		printf(" el %08lx of %08lx\n",
+		       (unsigned long) ((temp_ptr_int_t) *buf),
+		       (unsigned long) ((temp_ptr_int_t) ie));
 #endif
 		SEXP el = PROTECT(decode_to_SEXP(buf));
 		SEXP ea = R_NilValue;
 		if (ty == XT_LANG_TAG || ty==XT_LIST_TAG) {
 #ifdef RSERV_DEBUG
-		    printf(" tag %08lx of %08lx\n", (unsigned long)*buf, (unsigned long) ie);
+		    printf(" tag %08lx of %08lx\n",
+			   (unsigned long) ((temp_ptr_int_t) *buf),
+			   (unsigned long) ((temp_ptr_int_t) ie));
 #endif
 		    ea = decode_to_SEXP(buf);
 		    if (ea != R_NilValue) PROTECT(ea);
@@ -258,8 +272,8 @@ SEXP decode_to_SEXP(unsigned int **buf)
 		    UNPROTECT((ea == R_NilValue) ? 2 : 3);
 		} else {
 		    UNPROTECT((ea == R_NilValue) ? 2 : 3);
-		    PROTECT(val); /* protect the root */
 		    val = vnext;
+		    PROTECT(val); /* protect the root */
 		}
 		vtail = vnext;				   
 	    }
@@ -275,15 +289,16 @@ SEXP decode_to_SEXP(unsigned int **buf)
     if (vatt) {
 	/* if vatt contains "class" we have to set the object bit [we could use classgets(vec,kls) instead] */
 	SEXP head = vatt;
-	int has_class = 0;
+	SEXP has_class = 0;
 	PROTECT(val);
 	SET_ATTRIB(val, vatt);
 	while (head != R_NilValue) {
 	    if (TAG(head) == R_ClassSymbol) {
-		has_class = 1; break;
+		has_class = CAR(head); break;
 	    }
 	    head = CDR(head);
 	}
+#if (R_VERSION < R_Version(4,5,0))
 	if (has_class) /* if it has a class slot, we have to set the object bit */
 	    SET_OBJECT(val, 1);
 #ifdef SET_S4_OBJECT
@@ -292,6 +307,10 @@ SEXP decode_to_SEXP(unsigned int **buf)
 	   we can only flag "pure" S4 objects */
 	if (TYPEOF(val) == S4SXP)
 	    SET_S4_OBJECT(val);
+#endif
+#else /* SET_OBJECT() is not longer in the API so we have to set the class twice */
+	if (has_class)
+	    Rf_setAttrib(val, R_ClassSymbol, has_class);
 #endif
 	UNPROTECT(2); /* vatt + val */
     }
